BACKGROUND_MESSAGE_SUBJECT = "__BACKGROUND_MESSAGE__"
CHAT_SUBJECT = "__CHAT__"
TOOL_CHANGE_SUBJECT = "__TOOL_CALL__"
AGENT_FUNCTION_CALL_SUBJECT = "__AGENT_FUNCTION_CALL__"
ERROR_SUBJECT = "__ERROR__"
EVENT_SUBJECT = "__EVENT__"
NOTIFICATION_SUBJECT = "__NOTIFICATION__"
COMMAND_SUBJECT = "__COMMAND__"
SYSTEM_MESSAGE_SUBJECT = "__SYSTEM_MESSAGE__"
LOG_SUBJECT = "__LOG__"
CONFIG_SUBJECT = "__CONFIG__"

GLOBAL_CHANNEL = "__GLOBAL_CHANNEL__"
GLOBAL_SESSION_ID = "__GLOBAL__"
GLOBAL_EVENT_CHANNEL = "__GLOBAL_EVENT_CHANNEL__"
GLOBAL_NOTIFICATION_CHANNEL = "__GLOBAL_NOTIFICATION_CHANNEL__"


BACKGROUND_SERVICE_NAME = "background.service"
CONFIG_SERVICE_NAME = "config.service"
LOGGER_SERVICE_NAME = "logger.service"
NOTIFICATION_SERVICE_NAME = "notification.service"
OCR_SERVICE_NAME = "ocr.service"
VM_EVENT_MONITOR_SERVICE_NAME = "vm_event_monitoring.service"
REGISTRY_SERVICE_NAME = "registry.service"



import base64
from core_types.session_service import SessionService
from services.service_registry import ServiceRegistry
from common.constants import OCR_SERVICE_NAME
import easyocr
import cv2
import numpy as np
from typing import List
import logging

logger = logging.getLogger(__name__)


class OCRTextItem:
    text: str
    bbox: List[int]
    confidence: float

    def __init__(self, text: str, bbox: List[int], confidence: float):
        self.text = text
        self.bbox = bbox
        self.confidence = confidence


class OCRResult:
    text: List[OCRTextItem] = []
    average_confidence: float = 0.0
    strings: List[str] = []

    def __init__(self):
        self.text = []
        self._confidence = 0.0
        self.average_confidence = 0.0
        self.strings = []

    def add(self, text: str, bbox: List[int], confidence: float):
        if confidence > 0.5:
            self.text.append(OCRTextItem(text, bbox, confidence))
            self.strings.append(text)
        self._confidence += confidence
        self.average_confidence = self._confidence / len(self.text)


class OCRService(SessionService):
    def __init__(self, session_id: str, debug: bool = False):
        self.session_id = session_id
        self.debug = debug
        self.reader = easyocr.Reader(["en"])
        self.register()

    def name(self) -> str:
        return OCR_SERVICE_NAME

    def register(self) -> None:
        """Register this service with the registry"""
        if not ServiceRegistry.register(self.session_id, OCR_SERVICE_NAME, self):
            logger.info(f"OCR service already registered for session {self.session_id}")
            return

        logger.info(f"OCR service registered for session {self.session_id}")

    def unregister(self) -> None:
        """Unregister this service from the registry"""
        logger.info(f"OCR service unregistered for session {self.session_id}")

    def ocr(self, image_data: str) -> OCRResult:
        """OCR an image, the image should be a base64 encoded string"""
        logger.info(self.session_id, f"OCRing image: {image_data}")
        imageBytes = base64.b64decode(image_data)
        np_array = np.frombuffer(imageBytes, dtype=np.uint8)
        image = cv2.imdecode(np_array, cv2.IMREAD_COLOR)

        if len(image.shape) == 3:
            gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        else:
            gray = image

        results = self.reader.readtext(gray)
        result: OCRResult = OCRResult()
        for bbox, text, confidence in results:
            if self.debug:
                logger.info(
                    f"[{self.session_id}] Detected Text: {text} (Confidence: {confidence:.2f})"
                )
            result.add(text, bbox, confidence)
        return result



import cv2
import numpy as np
from collections import Counter
import base64
import logging

logger = logging.getLogger(__name__)


def get_dominant_border_color(
    image_data: str, border_percentage: float = 0.2, debug: bool = False
) -> str:
    """
    Detect the most prevalent color in the borders of an image.

    Args:
        image (np.ndarray): Input image (BGR format).
        border_percentage (float): Percentage of the border to consider (0.0 to 1.0).

    Returns:
        tuple: Dominant BGR color (B, G, R).
    """
    imageBytes = base64.b64decode(image_data)
    np_array = np.frombuffer(imageBytes, dtype=np.uint8)
    image = cv2.imdecode(np_array, cv2.IMREAD_COLOR)

    if image is None:
        return "#F2F2F7"

    if len(image.shape) == 2:  # Grayscale image
        image = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)

    height, width, _ = image.shape
    border_thickness_h = int(height * border_percentage)
    border_thickness_w = int(width * border_percentage)

    top_border = image[:border_thickness_h, :]
    bottom_border = image[-border_thickness_h:, :]
    left_border = image[:, :border_thickness_w]
    right_border = image[:, -border_thickness_w:]

    borders_combined = np.concatenate(
        (
            top_border.reshape(-1, 3),
            bottom_border.reshape(-1, 3),
            left_border.reshape(-1, 3),
            right_border.reshape(-1, 3),
        ),
        axis=0,
    )

    # Count the most common color
    pixel_counts = Counter(map(tuple, borders_combined))
    dominant_color = pixel_counts.most_common(1)[0][0]

    hex_color = "#%02X%02X%02X" % dominant_color

    if debug:
        logger.info(f"Dominant Border Color: {hex_color}")

    return hex_color


def detect_black_screen(
    image_data: str,
    black_threshold: int = 30,
    black_percentage: float = 0.9,
    debug: bool = False,
) -> bool:
    """
    Detect if an image is mostly black.

    Args:
        image (np.ndarray): The input image (BGR or grayscale).
        black_threshold (int): Pixel values below this are considered black (0-255).
        black_percentage (float): Percentage threshold to classify as mostly black (0.0 - 1.0).

    Returns:
        bool: True if the image is mostly black, False otherwise.
    """
    imageBytes = base64.b64decode(image_data)
    image = np.frombuffer(imageBytes, dtype=np.uint8)

    if len(image.shape) == 3:
        gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    else:
        gray = image

    black_pixels_mask = gray < black_threshold

    black_pixels_count = np.sum(black_pixels_mask)
    total_pixels = gray.size

    black_ratio = black_pixels_count / total_pixels

    if debug:
        logger.info(
            f"Black Pixels: {black_pixels_count}, Total Pixels: {total_pixels}, Black Ratio: {black_ratio:.2%}"
        )

    return bool(black_ratio >= black_percentage)



from typing import Optional, Dict, Any
from messages.message import Message
from websockets.legacy.server import WebSocketServerProtocol
import json
import uuid
from services.session_manager import SessionManager
from messages.error_message import create_error_message
from messages.event_message import create_event_message
from messages.error_message import create_error_message_from_message
import logging
import asyncio
import threading
from concurrent.futures import ThreadPoolExecutor
from queue import Queue
from common.constants import GLOBAL_CHANNEL, NOTIFICATION_SERVICE_NAME
from core_types.session_service import SessionService
from .service_registry import ServiceRegistry

logger = logging.getLogger(__name__)


class NotificationService(SessionService):
    def __init__(
        self,
        session_id: str,
        websocket: WebSocketServerProtocol,
        debug: bool = False,
    ):
        self._session_id = session_id
        self._session_manager = SessionManager.get_instance()
        self._connection = websocket
        self._debug = debug
        self._executor = ThreadPoolExecutor(max_workers=1)
        self._notification_queue: Queue = Queue()
        self._should_process = True
        self._main_loop = asyncio.get_event_loop()  # Store the main event loop
        self._start_queue_processor()
        self.register()

    def name(self) -> str:
        return NOTIFICATION_SERVICE_NAME

    def register(self) -> None:
        """Register this service with the registry"""
        if not ServiceRegistry.register(
            self._session_id, NOTIFICATION_SERVICE_NAME, self
        ):
            logger.info(
                f"Notification service already registered for session {self._session_id}"
            )
            return

    def unregister(self) -> None:
        """Unregister this service from the registry"""
        self.stop_queue_processor()
        logger.info(f"Notification service unregistered for session {self._session_id}")

    def update_websocket(self, websocket: WebSocketServerProtocol):
        """Update the websocket connection"""
        self._connection = websocket

    def _handle_send_result(self, future):
        """Handle the result of the send operation"""
        try:
            future.result()
        except Exception as e:
            logger.error(f"Error in send operation: {e}")

    def _start_queue_processor(self):
        """Start the background thread for processing notifications"""

        async def process_notifications_async():
            while self._should_process:
                try:
                    if not self._notification_queue.empty():
                        notification = (
                            self._notification_queue.get_nowait()
                        )  # Changed to non-blocking
                        try:
                            # Schedule the send operation on the main loop
                            future = asyncio.run_coroutine_threadsafe(
                                self.send(notification), self._main_loop
                            )
                            # Don't wait for the result, let it run independently
                            future.add_done_callback(
                                lambda f: self._handle_send_result(f)
                            )
                        except Exception as e:
                            logger.error(f"Error sending notification: {e}")
                    await asyncio.sleep(0.01)  # Reduced sleep time
                except asyncio.CancelledError:
                    break
                except Exception as e:
                    logger.error(f"Error in notification processor: {e}")

        def run_async_loop():
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            try:
                loop.run_until_complete(process_notifications_async())
            finally:
                loop.close()

        self._processor_thread = threading.Thread(target=run_async_loop, daemon=True)
        self._processor_thread.start()

    def stop_queue_processor(self):
        """Stop the queue processor thread"""
        self._should_process = False
        if hasattr(self, "_processor_thread"):
            self._processor_thread.join(timeout=5)  # Wait for thread to finish

    def queue_notification(self, message: Message):
        """Add a notification to the queue"""
        if not self._should_process:
            logger.warning("Queue processor is stopped, message will not be processed")
            return
        self._notification_queue.put(message)
        logger.debug(f"Queued notification: {message.to_dict()}")

    def send_sync(self, message: Message):
        """Queue a message for sending"""
        self.queue_notification(message)

    def send_error_sync(self, channel: Optional[str], error: str):
        """Queue an error message"""
        error_message = create_error_message(
            session_id=self._session_id,
            channel=channel if channel is not None else str(uuid.uuid4()),
            error_message=error,
        )
        self.queue_notification(error_message)

    def send_event_sync(
        self,
        channel: Optional[str],
        event: str,
        event_type: Optional[str],
        event_data: Dict[str, Any] | list[Dict[str, Any]] | None = None,
    ):
        """Queue an event message"""
        event_message = create_event_message(
            self._session_id, channel, event, event_type, event_data
        )
        self.queue_notification(event_message)

    async def send(self, message: Message) -> bool:
        """Send a message to a specific session"""
        if not self._session_manager.is_session_valid(message.session_id):
            return False

        websocket = self._connection
        if not websocket:
            return False

        try:
            dict_message = message.to_dict()
            await websocket.send(json.dumps(dict_message))
            if self._debug:
                logger.info(f"Sent message: {dict_message}")
            return True
        except Exception as e:
            logger.error(f"Error sending message: {e}")
            return False

    async def send_event(
        self,
        channel: Optional[str],
        event: str,
        event_type: Optional[str],
        event_data: Dict[str, Any] | list[Dict[str, Any]] | None = None,
    ):
        """Send an event message to a specific session"""
        event_message = create_event_message(
            self._session_id, channel, event, event_type, event_data
        )
        await self.send(event_message)
        if self._debug:
            logger.debug(f"Sent event message: {event_message.to_dict()}")

    async def send_error(self, channel: Optional[str], error: str):
        """Send an error message to a specific session"""
        error_message = create_error_message(
            session_id=self._session_id,
            channel=channel if channel is not None else str(uuid.uuid4()),
            error_message=error,
        )

        await self.send(error_message)
        if self._debug:
            logger.debug(f"Sent error message: {error_message.to_dict()}")

    async def send_exception(
        self,
        e: Exception,
        message: Message | None = None,
        channel: Optional[str] = None,
    ):
        """Send an error message to a specific session"""
        if message is not None:
            error_message = create_error_message_from_message(
                message, error_message=str(e)
            )
        else:
            error_message = create_error_message(
                session_id=self._session_id,
                channel=channel if channel is not None else str(uuid.uuid4()),
                error_message=str(e),
            )
        await self.send(error_message)
        if self._debug:
            logger.debug(f"Sent exception message: {error_message.to_dict()}")

    async def broadcast(self, message: Message):
        """Broadcast a message to all active sessions"""
        try:
            if self._session_manager.is_session_valid(self._session_id):
                broadcast_message = Message(
                    session_id=self._session_id,
                    channel=GLOBAL_CHANNEL,
                    subject=message.subject,
                    body=message.body,
                    context=message.context,
                )
                await self.send(broadcast_message)
                if self._debug:
                    logger.debug(
                        f"Broadcasted message {message.to_dict()} to session {self._session_id}"
                    )
        except Exception as e:
            logger.error(f"Error broadcasting to session {self._session_id}: {e}")

    def broadcast_sync(self, message: Message):
        """Broadcast a message to all active sessions"""
        broadcast_message = Message(
            session_id=self._session_id,
            channel=GLOBAL_CHANNEL,
            subject=message.subject,
            body=message.body,
            context=message.context,
        )
        self.send_sync(broadcast_message)
        if self._debug:
            logger.debug(
                f"Broadcasted message {message.to_dict()} to session {self._session_id}"
            )



import subprocess
from typing import List
import os
import base64
from helpers.image import get_dominant_border_color
from .models.get_vm_screenshot_result import GetVmScreenshotResult
from .helpers import get_prlctl_command


def get_vm_screenshot(vm_id: str) -> GetVmScreenshotResult:
    """Get a screenshot of the VM"""
    if not vm_id:
        raise ValueError("VM ID is required")
    try:
        temp_filename = f"{vm_id}_{os.urandom(4).hex()}.png"
        cmd: List[str] = [
            get_prlctl_command(),
            "capture",
            vm_id,
            "--file",
            temp_filename,
        ]
        result = subprocess.run(
            cmd,
            capture_output=True,
            text=True,
            check=False,
            shell=False,
        )
        if result.returncode != 0:
            return GetVmScreenshotResult(
                success=False,
                message=result.stderr,
                raw_screenshot=None,
                screenshot=None,
                border_color=None,
            )
        # Generate random temp filename to avoid collisions
        with open(temp_filename, "rb") as f:
            screenshot_data = f.read()
        os.remove(temp_filename)
        screenshot_data_str = _serialize_image(screenshot_data)
        border_color = get_dominant_border_color(screenshot_data_str)
        return GetVmScreenshotResult(
            success=True,
            message=f"Screenshot saved to {vm_id}.png",
            raw_screenshot=screenshot_data,
            screenshot=screenshot_data_str,
            border_color=border_color,
        )
    except Exception as e:
        return GetVmScreenshotResult(
            success=False,
            message=f"Failed to get screenshot for VM {vm_id}: {e}",
            raw_screenshot=None,
            screenshot=None,
            border_color=None,
        )


def _serialize_image(image_bytes: bytes) -> str:
    # Convert bytes to a base64 string
    base64_image = base64.b64encode(image_bytes).decode("utf-8")

    # JSON object with the base64 image
    return base64_image



from pydantic import BaseModel
from enum import Enum


class SetVmStateResult(BaseModel):
    success: bool
    message: str
    exit_code: int
    error: str
    vm: dict

    def to_dict(self) -> dict:
        return self.model_dump()


class VirtualMachineState(Enum):
    UNKNOWN = "unknown"
    START = "start"
    STOP = "stop"
    SUSPEND = "suspend"
    RESUME = "resume"
    PAUSE = "pause"
    SHUTDOWN = "shutdown"
    RESTART = "reboot"



from dataclasses import dataclass
from typing import Any, Dict
from core_types.content_message import ContentMessage
from common import SYSTEM_MESSAGE_SUBJECT
from .message import Message
from enum import Enum
import logging

logger = logging.getLogger(__name__)


class SystemMessageStatus(Enum):
    STARTED = "started"
    COMPLETED = "completed"
    FAILED = "failed"
    CANCELLED = "cancelled"
    SKIPPED = "skipped"
    PENDING = "pending"


@dataclass
class SystemMessage(ContentMessage):
    status: SystemMessageStatus
    link_id: str
    error: str
    message: str
    data: Dict[str, Any]

    def __init__(
        self,
        status: SystemMessageStatus,
        link_id: str,
        error: str,
        message: str,
        data: Dict[str, Any],
    ):
        self.status = status
        self.link_id = link_id
        self.error = error
        self.message = message
        self.data = data

    def subject(self) -> str:
        return SYSTEM_MESSAGE_SUBJECT

    def to_dict(self) -> Dict[str, Any]:
        return {
            "status": self.status.value,
            "link_id": self.link_id,
            "error": self.error,
            "message": self.message,
            "data": self.data,
        }

    def get(self, key: str) -> Any:
        return self.data.get(key)

    def copy(self) -> "SystemMessage":
        return SystemMessage(
            self.status, self.link_id, self.error, self.message, self.data
        )

    @staticmethod
    def from_dict(data: ContentMessage | dict[str, Any]) -> "SystemMessage | None":
        """Create a BackgroundMessage from a dictionary, returns None if invalid"""
        if isinstance(data, ContentMessage):
            data = data.to_dict()

        try:
            if not isinstance(data, dict):
                return None

            status = data.get("status")
            link_id = data.get("link_id")
            error = data.get("error")
            message = data.get("message")
            message_data = data.get("data", {})

            if not status:
                return None

            if status.upper() not in SystemMessageStatus.__members__:
                return None

            result = SystemMessage(
                status=SystemMessageStatus(status),
                link_id=link_id if link_id else "",
                error=error if error else "",
                message=message if message else "",
                data=message_data if message_data else {},
            )
            return result
        except Exception as e:
            logger.error(f"Error creating SystemMessage from dict: {e}")
            return None


def create_system_message(
    status: SystemMessageStatus,
    link_id: str,
    error: str,
    message: str,
    data: Dict[str, Any],
) -> SystemMessage:
    return SystemMessage(status, link_id, error, message, data)


def create_system_message_from_message(
    status: SystemMessageStatus,
    message: Message,
) -> Message:
    msg = Message(
        session_id=message.session_id,
        channel=message.channel,
        subject=SYSTEM_MESSAGE_SUBJECT,
        body=SystemMessage(
            status=status,
            link_id=message.message_id,
            error="",
            message="",
            data={},
        ).to_dict(),
    )
    msg.is_complete = message.is_complete
    msg.status = message.status
    msg.linked_message_id = message.linked_message_id
    msg.is_partial = message.is_partial
    return msg



from pydantic import BaseModel


class ExecuteVmCommandResult(BaseModel):
    output: str = ""
    error: str = ""
    exit_code: int = 0

    def __init__(self, output: str = "", error: str = "", exit_code: int = 0):
        self.output = output
        self.error = error
        self.exit_code = exit_code

    def to_dict(self):
        return {"output": self.output, "error": self.error, "exit_code": self.exit_code}

    @staticmethod
    def from_dict(data: dict):
        return ExecuteVmCommandResult(
            output=data.get("output", ""),
            error=data.get("error", ""),
            exit_code=data.get("exit_code", 0),
        )



from typing import List
import subprocess
from common.constants import NOTIFICATION_SERVICE_NAME
from services.service_registry import ServiceRegistry
from services.notification_service import NotificationService
from messages.message import Message
from messages.notification_message import (
    create_success_notification_message,
    create_error_notification_message,
)
from .helpers import get_prlsrvctl_command
from .models.operation_result import OperationResult
import logging

logger = logging.getLogger(__name__)


async def send_problem_report(
    session_id: str, vm_id: str, message: Message
) -> OperationResult:
    try:
        notifications = ServiceRegistry.get(
            session_id, NOTIFICATION_SERVICE_NAME, NotificationService
        )
        cmd: List[str] = [str(get_prlsrvctl_command()), "problem-report", "--send"]
        result = subprocess.run(
            cmd,
            capture_output=True,
            text=True,
            check=True,
            shell=False,
        )
        if result.returncode == 0:
            output = result.stdout.strip()
            report_id = None
            if "problem report was successfully sent with id:" in output.lower():
                report_id = output.split(":")[-1].strip()
            if report_id:
                success_message = create_success_notification_message(
                    session_id=session_id,
                    channel=vm_id,
                    message="Problem report sent successfully",
                    details=f"The problem report was sent successfully with id: {report_id}",
                )
                await notifications.send(success_message)
                return OperationResult(
                    success=True,
                    message="Problem report sent successfully",
                    exit_code=result.returncode,
                    error=result.stderr,
                )
        else:
            error_message = create_error_notification_message(
                session_id=session_id,
                channel=vm_id,
                message="Error sending problem report",
                details="The problem report was not sent successfully",
            )
            await notifications.send(error_message)
            return OperationResult(
                success=False,
                message="Error sending problem report",
                exit_code=result.returncode,
                error=result.stderr,
            )
    except Exception as e:
        logger.error(f"Error sending problem report: {e}")
        error_message = create_error_notification_message(
            session_id=session_id,
            channel=vm_id,
            message=f"Error sending problem report: {e}",
            details="The problem report was not sent successfully",
        )
        await notifications.send(error_message)
        return OperationResult(
            success=False,
            message=f"Error sending problem report: {e}",
            exit_code=1,
            error=str(e),
        )



from typing import Dict, List, Optional, ClassVar
import logging
from .models.virtual_machine import VirtualMachine
from datetime import datetime, timedelta
from .get_vms_from_prlctl import get_vms_from_prlctl
from .vm_parser import parse_vm_json

logger = logging.getLogger(__name__)


class VirtualMachineDataSource:
    """Singleton class to manage VirtualMachine data source"""

    _instance: ClassVar[Optional["VirtualMachineDataSource"]] = None

    @classmethod
    def initialize(cls) -> None:
        """Initialize the singleton instance"""
        if cls._instance is None:
            cls._instance = cls()

    @classmethod
    def get_instance(cls) -> "VirtualMachineDataSource":
        """Get the singleton instance"""
        if cls._instance is None:
            raise RuntimeError("VMDataSource not initialized")
        return cls._instance

    def __init__(self):
        if self._instance is not None:
            raise RuntimeError("Use get_instance() to access VMDataSource")

        self._vms: Dict[str, VirtualMachine] = {}  # vm_id -> VM
        self._last_update: Optional[datetime] = None
        self._cache_duration = timedelta(seconds=300)  # cache for 5 minutes
        vms = get_vms_from_prlctl()
        if vms.success:
            for vm in vms.raw_result:
                vm_model = parse_vm_json(vm)
                self._vms[vm_model.id] = vm_model

    def update_vm(self, vm: VirtualMachine) -> None:
        """Update a single VM in the cache"""
        self._vms[vm.id] = vm
        self._last_update = datetime.now()

    def update_vm_state(self, vm_id: str, state: str) -> None:
        """Update the state of a VM in the cache"""
        self._vms[vm_id].state = state
        self._last_update = datetime.now()

    def update_vms(self, vms: List[VirtualMachine]) -> None:
        """Update multiple VMs in the cache"""
        for vm in vms:
            self._vms[vm.id] = vm
        self._last_update = datetime.now()

    def get_vm(self, vm_id: str) -> Optional[VirtualMachine]:
        """Get a VirtualMachine by ID"""
        return self._vms.get(vm_id)

    def get_all_vms(self) -> List[VirtualMachine]:
        """Get all VMs"""
        return list(self._vms.values())

    def get_vms_by_state(self, state: str) -> List[VirtualMachine]:
        """Get all VMs in a specific state"""
        return [vm for vm in self._vms.values() if vm.state.lower() == state.lower()]

    def get_running_vms_count(self) -> int:
        """Get count of running VMs"""
        return len(self.get_vms_by_state("running"))

    def is_cache_valid(self) -> bool:
        """Check if the cache is still valid"""
        if self._last_update is None:
            return False
        return datetime.now() - self._last_update < self._cache_duration

    def clear_cache(self) -> None:
        """Clear the cache"""
        self._vms.clear()
        self._last_update = None

    def remove_vm(self, vm_id: str) -> None:
        """Remove a VM from the cache"""
        self._vms.pop(vm_id, None)

    def get_vms_by_name_pattern(self, pattern: str) -> List[VirtualMachine]:
        """Get VMs matching a name pattern"""
        return [vm for vm in self._vms.values() if pattern.lower() in vm.name.lower()]

    def get_vms_stats(self) -> Dict[str, int]:
        """Get VM statistics by state"""
        stats = {}
        for vm in self._vms.values():
            state = vm.state.lower()
            stats[state] = stats.get(state, 0) + 1
        return stats

    def length(self) -> int:
        """Get the number of VMs in the cache"""
        return len(self._vms)



from .get_vm_screenshot import get_vm_screenshot
from .models.get_vm_result import GetVmResult
from .datasource import VirtualMachineDataSource


def get_vms(take_screenshot: bool = False) -> GetVmResult:
    datasource = VirtualMachineDataSource.get_instance()
    vms = datasource.get_all_vms()
    if take_screenshot:
        for vm in vms:
            vm_screenshot_result = get_vm_screenshot(vm.id)
            vm.screenshot = vm_screenshot_result.screenshot
            datasource.update_vm(vm)

    return GetVmResult(
        success=True, message="VMs listed successfully", exit_code=0, error="", vms=vms
    )


def get_vm(vm_id: str, take_screenshot: bool = False) -> GetVmResult:
    datasource = VirtualMachineDataSource.get_instance()
    vm = datasource.get_vm(vm_id)
    if take_screenshot:
        vm_screenshot_result = get_vm_screenshot(vm.id)
        vm.screenshot = vm_screenshot_result.screenshot
        datasource.update_vm(vm)
    return GetVmResult(
        success=True, message="VM listed successfully", exit_code=0, error="", vm=vm
    )



from .helpers import get_prlctl_command
from typing import List
import subprocess
import json
from .get_vm_screenshot import get_vm_screenshot
from .models.get_vm_result import GetVmResult
from .vm_parser import parse_vm_json


def get_vms_from_prlctl(take_screenshot: bool = False) -> GetVmResult:
    try:
        cmd: List[str] = [str(get_prlctl_command()), "list", "-a", "-i", "--json"]
        result = subprocess.run(
            cmd,
            capture_output=True,
            text=True,
            check=True,
            shell=False,
        )
        data = json.loads(result.stdout)
        if take_screenshot:
            for vm in data:
                vm_screenshot_result = get_vm_screenshot(vm["ID"])
                vm["Screenshot"] = vm_screenshot_result.screenshot
        return GetVmResult(
            success=True,
            message="VMs listed successfully",
            exit_code=0,
            error="",
            raw_vms=data,
            vms=[parse_vm_json(vm) for vm in data],
        )
    except subprocess.CalledProcessError as e:
        return GetVmResult(
            success=False,
            message=f"Failed to list VMs: {e}",
            exit_code=1,
            error=str(e),
            raw_vms=[],
            vms=[],
        )
    except json.JSONDecodeError as e:
        return GetVmResult(
            success=False,
            message=f"Failed to parse VM list output: {e}",
            exit_code=1,
            error=str(e),
            raw_vms=[],
            vms=[],
        )


def get_vm_from_prlctl(vm_id: str, take_screenshot: bool = False) -> GetVmResult:
    if not vm_id:
        raise RuntimeError("No VM ID provided")
    try:
        cmd: List[str] = [
            str(get_prlctl_command()),
            "list",
            vm_id,
            "-a",
            "-i",
            "--json",
        ]
        result = subprocess.run(
            cmd,
            capture_output=True,
            text=True,
            check=True,
            shell=False,
        )
        result = json.loads(result.stdout)
        if len(result) == 0:
            raise RuntimeError(f"VM {vm_id} not found")
        if take_screenshot:
            for vm in result:
                if vm["ID"] == vm_id:
                    vm_screenshot_result = get_vm_screenshot(vm_id)
                    vm["Screenshot"] = vm_screenshot_result.screenshot
        return GetVmResult(
            success=True,
            message=f"VM {vm_id} listed successfully",
            exit_code=0,
            error="",
            raw_result=result[0],
            vm=parse_vm_json(result[0]),
        )
    except subprocess.CalledProcessError as e:
        return GetVmResult(
            success=False,
            message=f"Failed to get VM {vm_id} information: {e}",
            exit_code=1,
            error=str(e),
            raw_result={},
            vm=None,
        )
    except json.JSONDecodeError as e:
        return GetVmResult(
            success=False,
            message=f"Failed to parse VM {vm_id} information: {e}",
            exit_code=1,
            error=str(e),
            raw_result={},
            vm=None,
        )



from .message import Message
from core_types.content_message import ContentMessage
from typing import Dict, Any, Optional
import uuid
from common import (
    CHAT_SUBJECT,
    TOOL_CHANGE_SUBJECT,
    AGENT_FUNCTION_CALL_SUBJECT,
    GLOBAL_CHANNEL,
)
from common.message_status import MessageStatus


class StreamChatMessage(ContentMessage):
    def __init__(self, sender: str, role: str, content: str):
        self.sender = sender
        self.role = role
        self.content = content
        self.msg_type = "chat_stream"

    def to_dict(self) -> Dict[str, Any]:
        return {
            "sender": self.sender,
            "role": self.role,
            "content": self.content,
            "msg_type": self.msg_type,
        }

    def copy(self) -> "StreamChatMessage":
        return StreamChatMessage(self.sender, self.role, self.content)

    def subject(self) -> str:
        return CHAT_SUBJECT

    def get(self, key: Optional[str] = None) -> Any:
        if key == "content":
            return self.content
        if key == "sender":
            return self.sender
        if key == "role":
            return self.role
        if key == "msg_type":
            return self.msg_type
        return None


class ToolChangeChatMessage(ContentMessage):
    def __init__(self, name: str, arguments: Dict[str, Any]):
        self.name = name
        self.arguments = arguments
        self.msg_type = "tool_change"

    def to_dict(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "arguments": self.arguments,
            "msg_type": self.msg_type,
        }

    def copy(self) -> "ToolChangeChatMessage":
        return ToolChangeChatMessage(self.name, self.arguments)

    def subject(self) -> str:
        return TOOL_CHANGE_SUBJECT

    def get(self, key: Optional[str] = None) -> Any:
        if key == "name":
            return self.name
        if key == "arguments":
            return self.arguments
        if key == "msg_type":
            return self.msg_type
        if self.arguments is not None:
            for key in self.arguments:
                if key == key:
                    return self.arguments[key]
        return None


class AgentFunctionCallChatMessage(ContentMessage):
    def __init__(self, name: str, arguments: Dict[str, Any]):
        self.name = name
        self.arguments = arguments
        self.msg_type = "agent_function_call"

    def to_dict(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "arguments": self.arguments,
            "msg_type": self.msg_type,
        }

    def copy(self) -> "AgentFunctionCallChatMessage":
        return AgentFunctionCallChatMessage(self.name, self.arguments)

    def subject(self) -> str:
        return AGENT_FUNCTION_CALL_SUBJECT

    def get(self, key: Optional[str] = None) -> Any:
        if key == "name":
            return self.name
        if key == "arguments":
            return self.arguments
        if key == "msg_type":
            return self.msg_type
        if self.arguments is not None:
            for key in self.arguments:
                if key == key:
                    return self.arguments[key]
        return None


def create_stream_chat_message(
    session_id: str,
    channel: Optional[str],
    sender: str,
    role: str,
    content: str,
    is_complete: bool,
    linked_message_id: Optional[str] = None,
    is_partial: bool = False,
) -> Message:
    msg = Message(
        session_id=session_id,
        channel=channel if channel is not None else GLOBAL_CHANNEL,
        subject=CHAT_SUBJECT,
        body=StreamChatMessage(
            sender=sender,
            role=role,
            content=content,
        ),
    )
    msg.is_complete = is_complete
    msg.message_id = str(uuid.uuid4())
    if is_complete:
        msg.status = MessageStatus.COMPLETE
    else:
        msg.status = MessageStatus.STREAMING
    msg.linked_message_id = linked_message_id
    msg.is_partial = is_partial
    return msg


def create_tool_change_chat_message(
    session_id: str,
    channel: Optional[str],
    name: str,
    arguments: Dict[str, Any],
    linked_message_id: Optional[str] = None,
    is_partial: bool = False,
) -> Message:
    msg = Message(
        session_id=session_id,
        channel=channel if channel is not None else GLOBAL_CHANNEL,
        subject=TOOL_CHANGE_SUBJECT,
        body=ToolChangeChatMessage(
            name=name,
            arguments=arguments,
        ),
    )
    msg.is_complete = False
    msg.message_id = str(uuid.uuid4())
    msg.linked_message_id = linked_message_id
    msg.is_partial = is_partial
    return msg


def create_agent_function_call_chat_message(
    session_id: str,
    channel: Optional[str],
    name: str,
    arguments: Dict[str, Any] = {},
    linked_message_id: Optional[str] = None,
    is_partial: bool = False,
) -> Message:
    msg = Message(
        session_id=session_id,
        channel=channel if channel is not None else GLOBAL_CHANNEL,
        subject=AGENT_FUNCTION_CALL_SUBJECT,
        body=AgentFunctionCallChatMessage(
            name=name,
            arguments=arguments,
        ),
    )
    msg.is_complete = False
    msg.linked_message_id = linked_message_id
    msg.is_partial = is_partial
    msg.message_id = str(uuid.uuid4())
    return msg


def create_clean_agent_function_call_chat_message(
    session_id: str,
    channel: Optional[str],
    linked_message_id: Optional[str] = None,
    is_partial: bool = False,
) -> Message:
    msg = Message(
        session_id=session_id,
        channel=channel if channel is not None else GLOBAL_CHANNEL,
        subject=AGENT_FUNCTION_CALL_SUBJECT,
        body=AgentFunctionCallChatMessage(
            name="",
            arguments={},
        ),
    )
    msg.is_complete = False
    msg.linked_message_id = linked_message_id
    msg.is_partial = is_partial
    msg.message_id = str(uuid.uuid4())
    return msg



from typing import Dict, Any
from datetime import datetime
from core_types.llm_chat_ai_agent import LlmChatAgent
from enum import Enum


class ChannelStatus(Enum):
    ACTIVE = "active"
    CLOSED = "closed"
    ARCHIVED = "archived"


class SessionChannel:
    def __init__(self, channel_id: str, is_global: bool = False):
        self.id = channel_id
        self.created_at = datetime.now()
        self.last_active = datetime.now()
        self._messages: list = []
        self._metadata: Dict[str, Any] = {}
        self._context_variables: Dict[str, Any] = {}
        self._session_context: Dict[str, Any] = {}
        self.agent: LlmChatAgent | None = None
        self.is_global: bool = is_global
        self.status: ChannelStatus = ChannelStatus.ACTIVE

    def add_message(self, message: Dict[str, Any]) -> None:
        self._messages.append(message)
        self.update_last_active()

    def messages(self) -> list:
        return self._messages

    def clear_messages(self) -> None:
        self._messages = []

    def update_metadata(self, key: str, value: Any) -> None:
        self._metadata[key] = value

    def set_metadata(self, data: Dict[str, Any]) -> None:
        self._metadata.update(data)

    def metadata(self) -> Dict[str, Any]:
        return self._metadata

    def update_context_variable(self, key: str, value: Any) -> None:
        self._context_variables[key] = value

    def set_context_variables(self, data: Dict[str, Any]) -> None:
        self._context_variables.update(data)

    def context_variables(self) -> Dict[str, Any]:
        return self._context_variables

    def update_session_context(self, key: str, value: Any) -> None:
        self._session_context[key] = value

    def set_session_context(self, data: Dict[str, Any]) -> None:
        self._session_context.update(data)

    def session_context(self) -> Dict[str, Any]:
        return self._session_context

    def update_last_active(self) -> None:
        self.last_active = datetime.now()

    def close(self) -> None:
        """Close the channel"""
        self.status = ChannelStatus.CLOSED

    def archive(self) -> None:
        """Archive the channel"""
        self.status = ChannelStatus.ARCHIVED

    def to_dict(self) -> Dict[str, Any]:
        return {
            "id": self.id,
            "created_at": self.created_at.isoformat(),
            "last_active": self.last_active.isoformat(),
            "is_global": self.is_global,
            "status": self.status.value,
            "metadata": self._metadata,
            "context_variables": self._context_variables,
        }

    # ... other methods similar to current Session class



from typing import Dict, Any
from .models.virtual_machine import (
    VirtualMachine,
    GuestTools,
    MouseAndKeyboard,
    USBAndBluetooth,
    StartupAndShutdown,
    Optimization,
    TravelMode,
    Security,
    Expiration,
    Modality,
    Fullscreen,
    Coherence,
    TimeSynchronization,
    SMBIOSSettings,
    Hardware,
    CPU,
    Memory,
    Video,
    MemoryQuota,
    Hdd0,
    Cdrom0,
    Net0,
    Sound0,
    HostSharedFolders,
    SharedApplications,
    Network,
    MiscellaneousSharing,
    Advanced,
    SmartGuard,
    Usb,
    SharedProfile,
    SmartMount,
    IPAddresses,
)


def parse_vm_json(data: Dict[str, Any]) -> VirtualMachine:
    """Parse VM JSON data into a VM object"""

    # Parse nested objects first
    guest_tools = GuestTools(data=data)

    mouse_keyboard = MouseAndKeyboard(data=data)

    usb_bluetooth = USBAndBluetooth(data=data)

    startup_shutdown = StartupAndShutdown(data=data)

    optimization = Optimization(data=data)

    travel_mode = TravelMode(data=data)

    security = Security(data=data)

    smart_guard = SmartGuard(data=data)

    modality = Modality(data=data)

    fullscreen = Fullscreen(data=data)

    coherence = Coherence(data=data)

    time_sync = TimeSynchronization(data=data)

    expiration = Expiration(data=data)

    smbios = SMBIOSSettings(data=data)

    # Parse Hardware components
    cpu = CPU(data=data)

    memory = Memory(data=data)

    video = Video(data=data)

    memory_quota = MemoryQuota(data=data)

    hdd0 = Hdd0(data=data)

    cdrom0 = Cdrom0(data=data)

    usb = Usb(data=data)

    net0 = Net0(data=data)

    sound0 = Sound0(data=data)

    hardware = Hardware(
        cpu=cpu,
        memory=memory,
        video=video,
        memory_quota=memory_quota,
        hdd0=hdd0,
        cdrom0=cdrom0,
        usb=usb,
        net0=net0,
        sound0=sound0,
    )

    host_shared_folders = HostSharedFolders(data=data)

    shared_profile = SharedProfile(data=data)

    shared_applications = SharedApplications(data=data)

    smart_mount = SmartMount(data=data)

    ip_addresses = IPAddresses(data=data)

    network = Network(ip_addresses=ip_addresses)

    misc_sharing = MiscellaneousSharing(data=data)

    advanced = Advanced(data=data)

    # Create and return the VM object
    return VirtualMachine(
        id=data["ID"] if "ID" in data else "",
        name=data["Name"] if "Name" in data else "",
        description=data["Description"] if "Description" in data else "",
        type=data["Type"] if "Type" in data else "",
        state=data["State"] if "State" in data else "",
        os=data["OS"] if "OS" in data else "",
        template=data["Template"] if "Template" in data else "",
        uptime=int(data["Uptime"]) if "Uptime" in data else 0,
        home_path=data["Home path"] if "Home path" in data else "",
        home=data["Home"] if "Home" in data else "",
        restore_image=data["Restore Image"] if "Restore Image" in data else "",
        screenshot="",  # Not in sample JSON
        guest_tools=guest_tools,
        mouse_and_keyboard=mouse_keyboard,
        usb_and_bluetooth=usb_bluetooth,
        startup_and_shutdown=startup_shutdown,
        optimization=optimization,
        travel_mode=travel_mode,
        security=security,
        smart_guard=smart_guard,
        modality=modality,
        fullscreen=fullscreen,
        coherence=coherence,
        time_synchronization=time_sync,
        expiration=expiration,
        boot_order=data["Boot order"] if "Boot order" in data else "",
        bios_type=data["BIOS type"] if "BIOS type" in data else "",
        efi_secure_boot=data["EFI Secure boot"] if "EFI Secure boot" in data else "",
        allow_select_boot_device=(
            data["Allow select boot device"]
            if "Allow select boot device" in data
            else ""
        ),
        external_boot_device=(
            data["External boot device"] if "External boot device" in data else ""
        ),
        smbios_settings=smbios,
        hardware=hardware,
        host_shared_folders=host_shared_folders,
        host_defined_sharing=(
            data["Host defined sharing"] if "Host defined sharing" in data else ""
        ),
        shared_profile=shared_profile,
        shared_applications=shared_applications,
        smart_mount=smart_mount,
        network=network,
        miscellaneous_sharing=misc_sharing,
        advanced=advanced,
        print_management=None,  # Not in sample JSON
        guest_shared_folders=None,  # Not in sample JSON
    )



import threading
import subprocess
import time
import json
import logging
import selectors
from .service_registry import ServiceRegistry
from .notification_service import NotificationService
from .log_service import LogService
from common.constants import (
    GLOBAL_EVENT_CHANNEL,
    NOTIFICATION_SERVICE_NAME,
    LOGGER_SERVICE_NAME,
    VM_EVENT_MONITOR_SERVICE_NAME,
    BACKGROUND_SERVICE_NAME,
)
from events.prlctl_event_item import PrlctlEventItem
import os
import pty
from messages.event_message import create_event_message
from parallels_desktop.datasource import VirtualMachineDataSource
from parallels_desktop.get_vms import get_vms, get_vm
from parallels_desktop.vm_parser import parse_vm_json
from core_types.session_service import SessionService
from .background_service import BackgroundAgentService
from parallels_desktop.helpers import get_prlctl_command
from messages.constants import (
    VM_STATE_STARTED,
    VM_STATE_STOPPED,
    VM_STATE_SUSPENDED,
    VM_STATE_PAUSED,
    VM_STATE_CHANGED,
    VM_SYNC_SCREENSHOT,
)

logger = logging.getLogger(__name__)


class VmEventMonitorService(SessionService):
    def __init__(self, session_id: str, debug: bool = False):
        self._process = None
        self._running = False
        self._thread = None
        self._lock = threading.Lock()
        self.debug = debug
        self._session_id = session_id

        notifications_service = ServiceRegistry.get(
            session_id, NOTIFICATION_SERVICE_NAME, NotificationService
        )
        logger_service = ServiceRegistry.get(
            session_id, LOGGER_SERVICE_NAME, LogService
        )
        background_service = ServiceRegistry.get(
            session_id, BACKGROUND_SERVICE_NAME, BackgroundAgentService
        )
        if notifications_service is None:
            raise ValueError(f"Notification service not found for session {session_id}")
        if logger_service is None:
            raise ValueError(f"Logger service not found for session {session_id}")
        if background_service is None:
            raise ValueError(f"Background service not found for session {session_id}")
        self.notification_service = notifications_service
        self.logger_service = logger_service
        self.background_service = background_service
        self.datasource = VirtualMachineDataSource.get_instance()
        self.register()

    def name(self) -> str:
        return VM_EVENT_MONITOR_SERVICE_NAME

    def register(self) -> None:
        """Register this service with the registry"""
        if not ServiceRegistry.register(
            self._session_id, VM_EVENT_MONITOR_SERVICE_NAME, self
        ):
            logger.info(
                f"VM event monitor service already registered for session {self._session_id}"
            )
            return

        logger.info(
            f"VM event monitor service registered for session {self._session_id}"
        )

    def unregister(self) -> None:
        """Unregister this service from the registry"""
        self.stop()
        logger.info(
            f"VM event monitor service unregistered for session {self._session_id}"
        )

    def set_debug(self, debug: bool):
        self.debug = debug

    def start(self):
        """Start the VM event monitor service"""
        with self._lock:
            # Check if already running
            if self._running:
                logger.info("VM event monitor service is already running.")
                return

            # Double check process and thread status
            if self._process and self._process.poll() is None:
                logger.info("VM event monitor process is still running.")
                return

            if self._thread and self._thread.is_alive():
                logger.info("VM event monitor thread is still alive.")
                return

            logger.info("Starting VM event monitor service")
            self._running = True
            self._thread = threading.Thread(target=self._monitor_events, daemon=True)
            self._thread.start()

    def stop(self):
        """Stop the VM event monitor service"""
        with self._lock:
            self._running = False

            if self._process:
                logger.info("Terminating VM event monitor process")
                self._process.terminate()
                try:
                    self._process.wait(timeout=5)
                except subprocess.TimeoutExpired:
                    logger.warning("Force killing VM event monitor process")
                    self._process.kill()

            if self._thread:
                logger.info("Waiting for VM event monitor thread to finish")
                self._thread.join()

    def _monitor_events(self):
        """Monitor VM events in a background thread"""
        selector = selectors.DefaultSelector()

        while self._running:
            try:
                master_fd, slave_fd = pty.openpty()

                self._process = subprocess.Popen(
                    [get_prlctl_command(), "monitor-events", "--json"],
                    stdout=slave_fd,
                    stderr=subprocess.PIPE,
                    text=True,
                    bufsize=1,
                    close_fds=True,
                )

                os.close(slave_fd)
                logger.info("Started monitoring VM events")

                selector.register(master_fd, selectors.EVENT_READ)

                buffer = ""
                while self._running and self._process.poll() is None:
                    events = selector.select(timeout=1)
                    for key, _ in events:
                        if key.fileobj == master_fd:
                            try:
                                data = os.read(master_fd, 1024).decode(
                                    "utf-8", errors="replace"
                                )
                                if not data:
                                    break

                                buffer += data
                                while "\n" in buffer:
                                    line, buffer = buffer.split("\n", 1)
                                    line = line.strip()
                                    if line:
                                        if self.debug:
                                            logger.debug(f"STDOUT: {line}")
                                        self._handle_event(line)

                            except OSError as e:
                                logger.error(f"Error reading from PTY: {e}")
                                break

                selector.unregister(master_fd)
            except Exception as e:
                logger.exception(f"Error in VM event monitor: {e}")
                time.sleep(5)

            finally:
                if self._process and self._process.poll() is None:
                    self._process.terminate()
                    try:
                        self._process.wait(timeout=5)
                    except subprocess.TimeoutExpired:
                        self._process.kill()

                if self._running:
                    logger.warning(
                        "VM event monitor process exited unexpectedly. Restarting in 5 seconds..."
                    )
                    time.sleep(5)

    def _handle_event(self, line: str):
        """Handle a single event line from the VM monitor"""
        try:
            event_data = json.loads(line)
            event_item = PrlctlEventItem.from_dict(event_data)
            event_name = "vm_changed"
            event_type = "unknown"
            event_data = {}
            if event_item.vm_id:
                event_data["vm_id"] = event_item.vm_id
            if event_item.additional_info:
                for key, value in event_item.additional_info.items():
                    if (
                        key == "Vm state name"
                        and value != "cloning"
                        and value != "unknown"
                    ):
                        event_name = "state_changed"
                        event_type = value
                        self.datasource.update_vm_state(event_item.vm_id, value)
                        # notification_message = create_info_notification_message(
                        #     session_id=self._session_id,
                        #     channel=event_item.vm_id,
                        #     message=f"VM {event_item.vm_id} state changed to {value}",
                        # )
                        # self.notification_service.send_sync(notification_message)
                        # notification_message = create_info_notification_message(
                        #     session_id=self._session_id,
                        #     channel=GLOBAL_NOTIFICATION_CHANNEL,
                        #     message=f"VM {event_item.vm_id} state changed to {value}",
                        # )
                        # self.notification_service.send_sync(notification_message)
                    if event_type == "running":
                        self.background_service.post_message(
                            VM_SYNC_SCREENSHOT,
                            {"vm_id": event_item.vm_id},
                        )
                        self.background_service.post_message(
                            VM_STATE_STARTED,
                            {"vm_id": event_item.vm_id},
                        )
                        self.background_service.post_message(
                            VM_STATE_CHANGED,
                            {"vm_id": event_item.vm_id},
                        )
                    if event_type == "suspended":
                        self.background_service.post_message(
                            VM_STATE_SUSPENDED,
                            {"vm_id": event_item.vm_id},
                        )
                        self.background_service.post_message(
                            VM_STATE_CHANGED,
                            {"vm_id": event_item.vm_id},
                        )
                    if event_type == "stopped":
                        self.background_service.post_message(
                            VM_STATE_STOPPED,
                            {"vm_id": event_item.vm_id},
                        )
                        self.background_service.post_message(
                            VM_STATE_CHANGED,
                            {"vm_id": event_item.vm_id},
                        )
                    if event_type == "paused":
                        self.background_service.post_message(
                            VM_STATE_PAUSED,
                            {"vm_id": event_item.vm_id},
                        )
                        self.background_service.post_message(
                            VM_STATE_CHANGED,
                            {"vm_id": event_item.vm_id},
                        )
            if event_item.event_name == "vm_config_changed":
                event_name = "config_changed"
                event_type = "unknown"
            if event_item.event_name == "vm_snapshot_created":
                event_name = "snapshot_created"
                event_type = "unknown"
            if event_item.event_name == "vm_added":
                event_name = "vm_added"
                event_type = "vm_added"
                vms = get_vms()
                if vms.success:
                    for vm in vms.vm:
                        vm_model = parse_vm_json(vm)
                        self.datasource.update_vm(vm_model)
                new_vm = get_vm(event_item.vm_id)
                if new_vm.success:
                    event_data["vm"] = json.dumps(new_vm.vm)
            if event_item.event_name == "vm_unregistered":
                event_name = "vm_unregistered"
                event_type = "vm_unregistered"
                self.datasource.remove_vm(event_item.vm_id)
            event_message = create_event_message(
                session_id=self._session_id,
                channel=GLOBAL_EVENT_CHANNEL,
                event_name=event_name,
                event_type=event_type,
                event_data=event_data,
            )

            # only broadcast if event_type is known
            if event_type != "unknown":
                self.notification_service.broadcast_sync(event_message)

        except json.JSONDecodeError as e:
            logger.error(f"JSON decode error: {e} | Line: {line}")
        except Exception as e:
            logger.exception(f"Error handling VM event: {e}")



from pydantic import BaseModel


class OperationResult(BaseModel):
    success: bool
    message: str
    exit_code: int
    error: str

    def to_dict(self) -> dict:
        return self.model_dump()



import json
from typing import List, Callable, Union, Optional, Any, Dict
from common.defaults import DEFAULT_MODEL
from helpers.strings import normalize_string

# Third-party imports
from pydantic import BaseModel, HttpUrl

AgentFunction = Callable[[], Union[str, "LlmChatAgent", dict]]


class DataResult:
    value: Any
    context_variables: Dict[str, Any] = dict()


class LlmChatAgent:
    def __init__(
        self,
        name: str,
        instructions: Union[str, Callable[[Dict[str, Any]], str]],
        description: Optional[str] = None,
        model: str = DEFAULT_MODEL,
        functions: List[AgentFunction] = [],
        parallel_tool_calls: bool = True,
        transfer_instructions: Optional[str] = None,
        icon: Optional[Union[str, HttpUrl]] = None,
        tool_choice: Optional[str] = None,
    ):
        self.id = normalize_string(name)
        self.name = name
        self.description = description
        self.instructions = instructions
        self.model = model
        self.functions = functions or []
        self.parallel_tool_calls = parallel_tool_calls
        self.transfer_instructions = transfer_instructions
        self.icon = icon
        self.tool_choice = tool_choice


class LlmChatResponse(BaseModel):
    messages: List = []
    agent: Optional[LlmChatAgent] = None
    context_variables: dict = {}

    class Config:
        arbitrary_types_allowed = True


class LlmChatResult(BaseModel):
    """
    Encapsulates the possible return values for an agent function.

    Attributes:
        value (str): The result value as a string.
        agent (Dict): The agent instance as a dictionary.
        context_variables (dict): A dictionary of context variables.
    """

    value: str = ""
    agent: Optional[LlmChatAgent] = None
    context_variables: dict = {}

    class Config:
        arbitrary_types_allowed = True


class LlmChatAgentResponse:
    status: str
    message: str
    error: Optional[str] = None
    data: Optional[Union[dict, List[dict]]] = None
    agent: Optional[LlmChatAgent] = None
    context_variables: dict = {}

    def __init__(
        self,
        status: str,
        message: str,
        error: Optional[str] = None,
        data: Optional[Union[dict, List[dict]]] = None,
        agent: Optional[LlmChatAgent] = None,
        context_variables: dict = {},
    ):
        self.status = status
        self.message = message
        self.error = error
        self.data = data
        self.agent = agent
        self.context_variables = context_variables

    def to_dict(self):
        return {
            "status": self.status,
            "message": self.message,
            "error": self.error,
            "data": self.data if self.data is not None else None,
        }

    def value(self) -> str:
        if self.data is not None:
            return json.dumps(self.data)
        if self.error:
            return self.error
        if self.message:
            return self.message
        return self.status

    @staticmethod
    def from_dict(data: dict):
        return LlmChatAgentResponse(
            status=data["status"],
            message=data["message"],
            error=data["error"],
            data=data["data"],
        )
